"""
Module for handling Particle-Particle interaction.
"""

import numpy as np
from numba import njit

@njit
def update_0D(pos, id_ij, mass_ij, Lv, rc, potential_matrix, force, measure, rdf_hist):
    """
    Updates particles' accelerations when the cutoff radius :math: `r_c` is half the box's length, :math: `r_c = L/2`
    For no sub-cell. All ptcls within :math: `r_c = L/2` participate for force calculation. Cost ~ O(N^2)

    Parameters
    ----------
    force: func
        Force function.

    potential_matrix: array
        Potential parameters.

    rc: float
        Cut-off radius.

    Lv: array
        Array of box sides' lentgh.

    mass_ij: array
        Mass of each particle.

    id_ij: array
        Id of each particle

    pos: array
        Particles' positions.

    measure : bool
        Boolean for rdf calculation.

    rdf_hist : array
        Radial Distribution function array.

    Returns
    -------
    U_s_r : array
        Potential.

    acc_s_r : array
        Particles' accelerations.

    """
    L = Lv[0]
    Lh = L / 2.
    N = pos.shape[0]  # Number of particles

    U_s_r = 0.0  # Short-ranges potential energy accumulator
    acc_s_r = np.zeros_like(pos)  # Vector of accelerations

    rdf_nbins = rdf_hist.shape[0]
    dr_rdf = L / float(2.0 * rdf_nbins)

    for i in range(N):
        for j in range(i + 1, N):
            dx = (pos[i, 0] - pos[j, 0])
            dy = (pos[i, 1] - pos[j, 1])
            dz = (pos[i, 2] - pos[j, 2])

            if dx >= Lh:
                dx = L - dx
            elif dx <= -Lh:
                dx = L + dx

            if dy >= Lh:
                dy = L - dy
            elif dy <= -Lh:

                dy = L + dy

            if dz >= Lh:
                dz = L - dz
            elif dz <= -Lh:
                dz = L + dz

            # Compute distance between particles i and j
            r = np.sqrt(dx * dx + dy * dy + dz * dz)
            if measure and int(r / dr_rdf) < rdf_nbins:
                rdf_hist[int(r / dr_rdf), id_ij[i], id_ij[j]] += 1

            if 0 < r < rc:
                id_i = id_ij[i]
                id_j = id_ij[j]
                mass_i = mass_ij[i]
                mass_j = mass_ij[j]

                p_matrix = potential_matrix[:, id_i, id_j]
                # Compute the short-ranged force
                pot, fr = force(r, p_matrix)
                U_s_r += pot

                # Update the acceleration for i particles in each dimension

                acc_ix = dx * fr / mass_i
                acc_iy = dy * fr / mass_i
                acc_iz = dz * fr / mass_i

                acc_jx = dx * fr / mass_j
                acc_jy = dy * fr / mass_j
                acc_jz = dz * fr / mass_j

                acc_s_r[i, 0] += acc_ix
                acc_s_r[i, 1] += acc_iy
                acc_s_r[i, 2] += acc_iz

                # Apply Newton's 3rd law to update acceleration on j particles
                acc_s_r[j, 0] -= acc_jx
                acc_s_r[j, 1] -= acc_jy
                acc_s_r[j, 2] -= acc_jz

    return U_s_r, acc_s_r


@njit
def update(pos, p_id, p_mass, box_lengths, rc, potential_matrix, force, measure, rdf_hist):
    """
    Update the force on the particles based on a linked cell-list (LCL) algorithm.

    Parameters
    ----------
    force: float, float
        Potential and force values.

    potential_matrix: array
        Potential parameters.

    rc: float
        Cut-off radius.

    box_lengths: array
        Array of box sides' length.

    p_mass: array
        Mass of each particle.

    p_id: array
        Id of each particle

    pos: array
        Particles' positions.

    measure : bool
        Boolean for rdf calculation.

    rdf_hist : array
        Radial Distribution function array.

    Returns
    -------
    U_s_r : float
        Short-ranged component of the potential energy of the system.

    acc_s_r : array
        Short-ranged component of the acceleration for the particles.

    Notes
    -----
    Here the "short-ranged component" refers to the Ewald decomposition of the
    short and long ranged interactions. See the wikipedia article:
    https://en.wikipedia.org/wiki/Ewald_summation or
    "Computer Simulation of Liquids by Allen and Tildesley" for more information.
    """
    acc_s_r = np.zeros_like(pos)

    # Declare parameters
    N = pos.shape[0]  # Number of particles
    d = pos.shape[1]  # Number of dimensions
    rshift = np.zeros(d)  # Shifts for array flattening

    # Initialize
    U_s_r = 0.0  # Short-ranges potential energy accumulator
    ls = np.arange(N)  # List of particle indices in a given cell

    # The number of cells in each dimension
    cells_per_dim = (box_lengths / rc).astype(np.int64)
    cell_length_per_dim = box_lengths / cells_per_dim

    # Total number of cells in volume
    Ncell = cells_per_dim.prod()
    head = np.arange(Ncell)  # List of head particles
    empty = -50  # value for empty list and head arrays
    head.fill(empty)  # Make head list empty until population

    rdf_nbins = rdf_hist.shape[0]
    dr_rdf = rc / float(rdf_nbins)

    # Loop over all particles and place them in cells
    for i in range(N):
        # Determine what cell, in each direction, the i-th particle is in
        cx = int(pos[i, 0] / cell_length_per_dim[0])  # X cell
        cy = int(pos[i, 1] / cell_length_per_dim[1])  # Y cell
        cz = int(pos[i, 2] / cell_length_per_dim[2])  # Z cell

        # Determine cell in 3D volume for i-th particle
        c = cx + cy * cells_per_dim[0] + cz * cells_per_dim[0] * cells_per_dim[1]
        # List of particle indices occupying a given cell
        ls[i] = head[c]

        # The last particle found to lie in cell c (head particle)
        head[c] = i

    # Loop over all cells in x, y, and z direction
    for cx in range(cells_per_dim[0]):
        for cy in range(cells_per_dim[1]):
            for cz in range(cells_per_dim[2]):

                # Compute the cell in 3D volume
                c = cx + cy * cells_per_dim[0] + cz * cells_per_dim[0] * cells_per_dim[1]

                # Loop over all cell pairs (N-1 and N+1)
                for cz_N in range(cz - 1, cz + 2):
                    # z cells
                    # Check periodicity: needed for 0th cell
                    # if cz_N < 0:
                    #     cz_shift = cells_per_dim[2]
                    #     rshift[2] = -box_lengths[2]
                    # # Check periodicity: needed for Nth cell
                    # elif cz_N >= cells_per_dim[2]:
                    #     cz_shift = -cells_per_dim[2]
                    #     rshift[2] = box_lengths[2]
                    # else:
                    #     cz_shift = 0
                    #     rshift[2] = 0.0
                    cz_shift = 0 + cells_per_dim[2] * (cz_N < 0) - cells_per_dim[2] * (cz_N >= cells_per_dim[2])
                    rshift[2] = 0.0 - box_lengths[2] * (cz_N < 0) + box_lengths[2]*(cz_N >= cells_per_dim[2])

                    for cy_N in range(cy - 1, cy + 2):
                        # y cells
                        # Check periodicity
                        # if cy_N < 0:
                        #     cy_shift = cells_per_dim[1]
                        #     rshift[1] = -box_lengths[1]
                        # elif cy_N >= cells_per_dim[1]:
                        #     cy_shift = -cells_per_dim[1]
                        #     rshift[1] = box_lengths[1]
                        # else:
                        #     cy_shift = 0
                        #     rshift[1] = 0.0

                        cy_shift = 0 + cells_per_dim[1] * (cy_N < 0) - cells_per_dim[1] * (cy_N >= cells_per_dim[1])
                        rshift[1] = 0.0 - box_lengths[1] * (cy_N < 0) + box_lengths[1] * (cy_N >= cells_per_dim[1])

                        for cx_N in range(cx - 1, cx + 2):
                            # x cells
                            # Check periodicity
                            # if cx_N < 0:
                            #     cx_shift = cells_per_dim[0]
                            #     rshift[0] = -box_lengths[0]
                            # elif cx_N >= cells_per_dim[0]:
                            #     cx_shift = -cells_per_dim[0]
                            #     rshift[0] = box_lengths[0]
                            # else:
                            #     cx_shift = 0
                            #     rshift[0] = 0.0

                            cx_shift = 0 + cells_per_dim[0] * (cx_N < 0) - cells_per_dim[0] * (cx_N >= cells_per_dim[0])
                            rshift[0] = 0.0 - box_lengths[0] * (cx_N < 0) + box_lengths[0] * (cx_N >= cells_per_dim[0])

                            # Compute the location of the N-th cell based on shifts
                            c_N = (cx_N + cx_shift) + (cy_N + cy_shift) * cells_per_dim[0] \
                                  + (cz_N + cz_shift) * cells_per_dim[0] * cells_per_dim[1]

                            i = head[c]
                            # First compute interaction of head particle with neighboring cell head particles
                            # Then compute interactions of head particle within a specific cell
                            while i != empty:

                                # Check neighboring head particle interactions
                                j = head[c_N]

                                while j != empty:

                                    # Only compute particles beyond i-th particle (Newton's 3rd Law)
                                    if i < j:

                                        # Compute the difference in positions for the i-th and j-th particles
                                        dx = pos[i, 0] - (pos[j, 0] + rshift[0])
                                        dy = pos[i, 1] - (pos[j, 1] + rshift[1])
                                        dz = pos[i, 2] - (pos[j, 2] + rshift[2])

                                        # Compute distance between particles i and j
                                        r = np.sqrt(dx ** 2 + dy ** 2 + dz ** 2)

                                        if measure and int(r / dr_rdf) < rdf_nbins:
                                            rdf_hist[int(r / dr_rdf), p_id[i], p_id[j]] += 1

                                        # If below the cutoff radius, compute the force
                                        if r < rc:
                                            p_matrix = potential_matrix[:, p_id[i], p_id[j]]

                                            # Compute the short-ranged force
                                            pot, fr = force(r, p_matrix)
                                            fr /= r
                                            U_s_r += pot

                                            # Update the acceleration for i particles in each dimension

                                            acc_s_r[i, 0] += dx * fr / p_mass[i]
                                            acc_s_r[i, 1] += dy * fr / p_mass[i]
                                            acc_s_r[i, 2] += dz * fr / p_mass[i]

                                            # Apply Newton's 3rd law to update acceleration on j particles
                                            acc_s_r[j, 0] -= dx * fr / p_mass[j]
                                            acc_s_r[j, 1] -= dy * fr / p_mass[j]
                                            acc_s_r[j, 2] -= dz * fr / p_mass[j]

                                    # Move down list (ls) of particles for cell interactions with a head particle
                                    j = ls[j]

                                # Check if head particle interacts with other cells
                                i = ls[i]
    return U_s_r, acc_s_r



"""
Module handling stages of an MD run: PreProcessing, Simulation, PostProcessing.
"""
import numpy as np
import copy as py_copy
from numba import njit
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

import os

# Sarkas modules
from sarkas.utilities.io import InputOutput
from sarkas.utilities.timing import SarkasTimer
from sarkas.potentials.core import Potential
from sarkas.time_evolution.integrators import Integrator
from sarkas.time_evolution.thermostats import Thermostat
from sarkas.core import Particles, Parameters, Species
import sarkas.tools.observables as sk_obs


class Process:
    """Stage of a Molecular Dynamics simulation. This is the Parent class for PreProcess, Simulation, and PostProcess.

    Parameters
    ----------
    input_file : str
        Path to the YAML input file.

    Attributes
    ----------
    potential : sarkas.potential.base.Potential
        Class handling the interaction between particles.

    integrator: sarkas.time_evolution.integrators.Integrator
        Class handling the integrator.

    thermostat: sarkas.time_evolution.thermostats.Thermostat
        Class handling the equilibration thermostat.

    particles: sarkas.core.Particles
        Class handling particles properties.

    parameters: sarkas.core.Parameters
        Class handling simulation's parameters.

    species: list
        List of :meth:`sarkas.core.Species` classes.

    input_file: str
        Path to YAML input file.

    timer: sarkas.utilities.timing.SarkasTimer
        Class handling the timing of processes.

    io: sarkas.utilities.io.InputOutput
        Class handling the IO in Sarkas.

    """

    def __init__(self, input_file: str = None):
        self.potential = Potential()
        self.integrator = Integrator()
        self.thermostat = Thermostat()
        self.parameters = Parameters()
        self.particles = Particles()
        self.species = []
        self.input_file = input_file if input_file else None
        self.timer = SarkasTimer()
        self.io = InputOutput(process=self.__name__)

    def common_parser(self, filename: str = None) -> None:
        """
        Parse simulation parameters from YAML file.

        Parameters
        ----------
        filename: str
            Input YAML file

        """
        if filename:
            self.input_file = filename

        dics = self.io.from_yaml(self.input_file)

        for lkey in dics:
            if lkey == "Particles":
                for species in dics["Particles"]:
                    spec = Species(species["Species"])
                    self.species.append(spec)

            if lkey == "Potential":
                self.potential.from_dict(dics[lkey])

            if lkey == "Thermostat":
                self.thermostat.from_dict(dics[lkey])

            if lkey == "Integrator":
                self.integrator.from_dict(dics[lkey])

            if lkey == "Parameters":
                self.parameters.from_dict(dics[lkey])

        self.observables_list = []
        # This is not needed in the case of process = simulation
        for observable in dics['Observables']:
            for key, sub_dict in observable.items():
                if key == 'RadialDistributionFunction':
                    self.observables_list.append('rdf')
                    self.rdf = sk_obs.RadialDistributionFunction()
                    if sub_dict:
                        self.rdf.from_dict(sub_dict)
                if key == 'HermiteCoefficients':
                    self.hc = sk_obs.HermiteCoefficients()
                    self.hc.from_dict(sub_dict)
                if key == 'Thermodynamics':
                    self.therm = sk_obs.Thermodynamics()
                    self.therm.from_dict(sub_dict)
                    self.observables_list.append('therm')
                if key == 'DynamicStructureFactor':
                    self.observables_list.append('dsf')
                    self.dsf = sk_obs.DynamicStructureFactor()
                    if sub_dict:
                        self.dsf.from_dict(sub_dict)
                if key == 'CurrentCorrelationFunction':
                    self.observables_list.append('ccf')
                    self.ccf = sk_obs.CurrentCorrelationFunction()
                    if sub_dict:
                        self.ccf.from_dict(sub_dict)
                if key == 'StaticStructureFactor':
                    self.observables_list.append('ssf')
                    self.ssf = sk_obs.StaticStructureFactor()
                    if sub_dict:
                        self.ssf.from_dict(sub_dict)
                if key == 'VelocityAutoCorrelationFunction':
                    self.observables_list.append('vacf')
                    self.vacf = sk_obs.VelocityAutoCorrelationFunction()
                    if sub_dict:
                        self.vacf.from_dict(sub_dict)
                if key == 'VelocityDistribution':
                    self.observables_list.append('vd')
                    self.vm = sk_obs.VelocityDistribution()
                    if sub_dict:
                        self.vm.from_dict(sub_dict)
                if key == 'ElectricCurrent':
                    self.observables_list.append('ec')
                    self.ec = sk_obs.ElectricCurrent()
                    if sub_dict:
                        self.ec.from_dict(sub_dict)
                if key == 'DiffusionFlux':
                    self.observables_list.append('diff_flux')
                    self.diff_flux = sk_obs.DiffusionFlux()
                    if sub_dict:
                        self.diff_flux.from_dict(sub_dict)

        if 'TransportCoefficients' in dics.keys():
            self.transport_dict = dics["TransportCoefficients"].copy()

    def initialization(self):
        """Initialize all classes."""

        # initialize the directories and filenames
        self.io.setup()

        # Copy relevant subsclasses attributes into parameters class. This is needed for post-processing.

        # Update parameters' dictionary with filenames and directories
        self.parameters.from_dict(self.io.__dict__)
        # save some general info
        self.parameters.potential_type = self.potential.type
        self.parameters.cutoff_radius = self.potential.rc
        self.parameters.integrator = self.integrator.type
        self.parameters.thermostat = self.thermostat.type

        # Copy some integrator parameters if not already defined
        if not hasattr(self.parameters, 'dt'):
            self.parameters.dt = self.integrator.dt
        if not hasattr(self.parameters, 'equilibration_steps'):
            self.parameters.equilibration_steps = self.integrator.equilibration_steps
        if not hasattr(self.parameters, 'eq_dump_step'):
            self.parameters.eq_dump_step = self.integrator.eq_dump_step
        if not hasattr(self.parameters, 'production_steps'):
            self.parameters.production_steps = self.integrator.production_steps
        if not hasattr(self.parameters, 'prod_dump_step'):
            self.parameters.prod_dump_step = self.integrator.prod_dump_step

        # Check for magnetization phase
        if self.integrator.electrostatic_equilibration:
            self.parameters.electrostatic_equilibration = True
            if not hasattr(self.parameters, 'mag_dump_step'):
                self.parameters.mag_dump_step = self.integrator.mag_dump_step
            if not hasattr(self.parameters, 'magnetization_steps'):
                self.parameters.magnetization_steps = self.integrator.magnetization_steps

        self.parameters.setup(self.species)

        t0 = self.timer.current()
        self.potential.setup(self.parameters)
        time_pot = self.timer.current()

        self.thermostat.setup(self.parameters)
        self.integrator.setup(self.parameters, self.thermostat, self.potential)
        self.particles.setup(self.parameters, self.species)
        time_ptcls = self.timer.current()

        # For restart and backups.
        self.io.setup_checkpoint(self.parameters, self.species)
        self.io.save_pickle(self)

        # Print Process summary to file and screen
        self.io.simulation_summary(self)
        time_end = self.timer.current()

        # Print timing
        self.io.time_stamp("Potential Initialization", self.timer.time_division(time_end - t0))
        self.io.time_stamp("Particles Initialization", self.timer.time_division(time_ptcls - time_pot))
        self.io.time_stamp("Total Simulation Initialization", self.timer.time_division(time_end - t0))

    def setup(self, read_yaml=False, other_inputs=None):
        """Setup simulations' parameters and io subclasses.

        Parameters
        ----------
        read_yaml: bool
            Flag for reading YAML input file. Default = False.

        other_inputs: dict (optional)
            Dictionary with additional simulations options.

        """
        if read_yaml:
            self.common_parser()

        if other_inputs:
            assert isinstance(other_inputs, dict), "Wrong input type. other_inputs should be a nested dictionary"

            for class_name, class_attr in other_inputs.items():
                if class_name not in ['Particles', 'Obervables']:
                    self.__dict__[class_name.lower()].__dict__.update(class_attr)
                else:
                    for sp, species in enumerate(other_inputs["Particles"]):
                        spec = Species(species["Species"])
                        self.species[sp].__dict__.update(spec.__dict__)

                if class_name == 'Observables':

                    for observable in class_attr:
                        for key, sub_dict in observable.items():
                            if key == 'RadialDistributionFunction':
                                self.rdf = sk_obs.RadialDistributionFunction()
                                self.rdf.from_dict(sub_dict)
                            if key == 'HermiteCoefficients':
                                self.hc = sk_obs.HermiteCoefficients()
                                self.hc.from_dict(sub_dict)
                            if key == 'Thermodynamics':
                                self.therm = sk_obs.Thermodynamics()
                                self.therm.from_dict(sub_dict)
                            if key == 'DynamicStructureFactor':
                                self.dsf = sk_obs.DynamicStructureFactor()
                                if sub_dict:
                                    self.dsf.from_dict(sub_dict)
                            if key == 'CurrentCorrelationFunction':
                                self.ccf = sk_obs.CurrentCorrelationFunction()
                                if sub_dict:
                                    self.ccf.from_dict(sub_dict)
                            if key == 'StaticStructureFactor':
                                self.ssf = sk_obs.StaticStructureFactor()
                                if sub_dict:
                                    self.ssf.from_dict(sub_dict)
                            if key == 'VelocityAutoCorrelationFunction':
                                self.vacf = sk_obs.VelocityAutoCorrelationFunction()
                                if sub_dict:
                                    self.vacf.from_dict(sub_dict)
                            if key == 'VelocityMoments':
                                self.vm = sk_obs.VelocityMoments()
                                if sub_dict:
                                    self.vm.from_dict(sub_dict)
                            if key == 'ElectricCurrent':
                                self.ec = sk_obs.ElectricCurrent()
                                if sub_dict:
                                    self.ec.from_dict(sub_dict)

        if self.__name__ == 'postprocessing':

            # Create the file paths without creating directories and redefining io attributes
            self.io.create_file_paths()

            # Read previouly stored files
            self.io.read_pickle(self)

            # Print parameters to log file
            self.io.simulation_summary(self)

            # Initialize the observable classes
            # for obs in self.observables_list:
            #     if obs in self.__dict__.keys():
            #         self.__dict__[obs].setup(self.parameters)

        else:
            self.initialization()

        if self.parameters.plot_style:
            plt.style.use(self.parameters.plot_style)


class PostProcess(Process):
    """
    Class handling the post-processing stage of a simulation.

    Parameters
    ----------
    input_file : str
        Path to the YAML input file.

    """

    def __init__(self, input_file: str = None):
        self.__name__ = 'postprocessing'
        super().__init__(input_file)

    def setup_from_simulation(self, simulation):
        """
        Setup postprocess' subclasses by (shallow) copying them from simulation object.

        Parameters
        ----------
        simulation: sarkas.core.processes.Simulation
            Simulation object

        """
        self.parameters = py_copy.copy(simulation.parameters)
        self.integrator = py_copy.copy(simulation.integrator)
        self.potential = py_copy.copy(simulation.potential)
        self.species = py_copy.copy(simulation.species)
        self.thermostat = py_copy.copy(simulation.thermostat)
        self.io = py_copy.copy(simulation.io)

    def run(self):
        """Calculate all the observables from the YAML input file."""

        for obs in self.observables_list:
            if obs in self.__dict__.keys():
                self.__dict__[obs].setup(self.parameters)
                if obs == 'therm':
                    self.therm.temp_energy_plot(self)
                else:
                    self.io.postprocess_info(self, write_to_file=True, observable=obs)
                    self.__dict__[obs].compute()

        if hasattr(self, 'transport_dict'):
            from sarkas.tools.transport import TransportCoefficient as TC

            for coeff in self.transport_dict:

                for key, coeff_kwargs in coeff.items():

                    if key.lower() == 'diffusion':
                        TC.diffusion(self.parameters, **coeff_kwargs)

                    if key.lower() == 'interdiffusion':
                        TC.interdiffusion(self.parameters, **coeff_kwargs)

                    if key.lower() == 'viscosity':
                        TC.viscosity(self.parameters, **coeff_kwargs)

                    if key.lower() == 'electricalconductivity':
                        TC.electrical_conductivity(self.parameters, **coeff_kwargs)


class PreProcess(Process):
    """
    Wrapper class handling the estimation of time and best parameters of a simulation.

    Parameters
    ----------
    input_file : str
        Path to the YAML input file.

    Attributes
    ----------
    loops: int
        Number of timesteps to run for time and size estimates. Default = 10

    estimate: bool
        Run an estimate for the best PPPM parameters in the simulation. Default=False.

    pm_meshes: numpy.ndarray
        Array of mesh sizes used in the PPPM parameters estimation.

    pp_cells: numpy.ndarray
        Array of simulations box cells used in the PPPM parameters estimation.

    kappa: float
        Screening parameter. Calculated from :meth:`sarkas.potentials.core.Potential.matrix`.

    """

    def __init__(self, input_file: str = None):
        self.__name__ = 'preprocessing'
        self.loops = 10
        self.estimate = False
        self.pm_meshes = np.logspace(3, 7, 12, base =2, dtype=int )
        # np.array([16, 24, 32, 48, 56, 64, 72, 88, 96, 112, 128], dtype=int)
        self.pp_cells = np.arange(3, 16, dtype=int)
        self.kappa = None
        super().__init__(input_file)

    def green_function_timer(self):
        """Time Potential setup."""

        self.timer.start()
        self.potential.pppm_setup(self.parameters)

        return self.timer.stop()

    def run(self,
            loops: int = None,
            timing: bool = True,
            timing_study: bool = False,
            pppm_estimate: bool = False,
            postprocessing: bool = False,
            remove: bool = False):
        """
        Estimate the time of the simulation and best parameters if wanted.

        Parameters
        ----------
        loops : int
            Number of loops over which to average the acceleration calculation.
            Note that the number of timestep over which to averages is three times this value.
            Example: loops = 5, acceleration is averaged over 5 loops, while full time step over 15 loops.

        timing : bool
            Flag for estimating simulation times. Default =True.

        timing_study : bool
            Flag for estimating time for simulation parameters.

        pppm_estimate : bool
            Flag for showing the force error plots in case of pppm algorithm.

        postprocessing : bool
            Flag for calculating Post processing parameters.

        remove : bool
            Flag for removing energy files and dumps created during times estimation. Default = False.

        """

        plt.close('all')
        self.pppm_plots_dir = os.path.join(self.io.preprocessing_dir, 'PPPM_Plots')
        if not os.path.exists(self.pppm_plots_dir):
            os.mkdir(self.pppm_plots_dir)

        # Set the screening parameter
        self.kappa = self.potential.matrix[1, 0, 0] if self.potential.type == "Yukawa" else 0.0

        if loops:
            self.loops = loops + 1

        if timing:
            self.io.preprocess_timing("header", [0, 0, 0, 0, 0, 0], 0)
            if self.potential.pppm_on:
                green_time = self.timer.time_division(self.green_function_timer())
                self.io.preprocess_timing("GF", green_time, 0)

            self.time_acceleration()
            self.time_integrator_loop()

            # Estimate size of dump folder
            # Grab one file from the dump directory and get the size of it.
            eq_dump_size = os.stat(os.path.join(self.io.eq_dump_dir, os.listdir(self.io.eq_dump_dir)[0])).st_size
            eq_dump_fldr_size = eq_dump_size * (self.integrator.equilibration_steps / self.integrator.eq_dump_step)
            # Grab one file from the dump directory and get the size of it.
            prod_dump_size = os.stat(os.path.join(self.io.eq_dump_dir, os.listdir(self.io.eq_dump_dir)[0])).st_size
            prod_dump_fldr_size = prod_dump_size * (self.integrator.production_steps / self.integrator.prod_dump_step)
            # Prepare arguments to pass for print out
            sizes = np.array([[eq_dump_size, eq_dump_fldr_size],
                              [prod_dump_size, prod_dump_fldr_size]])
            # Check for electrostatic equilibration
            if self.integrator.electrostatic_equilibration:
                dump = self.integrator.mag_dump_step
                mag_dump_size = os.stat(os.path.join(self.io.mag_dump_dir, 'checkpoint_' + str(dump) + '.npz')).st_size
                mag_dump_fldr_size = mag_dump_size * (
                        self.integrator.magnetization_steps / self.integrator.mag_dump_step)
                sizes = np.array([[eq_dump_size, eq_dump_fldr_size],
                                  [prod_dump_size, prod_dump_fldr_size],
                                  [mag_dump_size, mag_dump_fldr_size]])

            self.io.preprocess_sizing(sizes)

            if remove:
                # Delete the energy files created during the estimation runs
                os.remove(self.io.eq_energy_filename)
                os.remove(self.io.prod_energy_filename)

                # Delete dumps created during the estimation runs
                for npz in os.listdir(self.io.eq_dump_dir):
                    os.remove(os.path.join(self.io.eq_dump_dir, npz))

                for npz in os.listdir(self.io.prod_dump_dir):
                    os.remove(os.path.join(self.io.prod_dump_dir, npz))

                if self.integrator.electrostatic_equilibration:
                    os.remove(self.io.mag_energy_filename)
                    # Remove dumps
                    for npz in os.listdir(self.io.mag_dump_dir):
                        os.remove(os.path.join(self.io.mag_dump_dir, npz))

        if pppm_estimate:
            if timing_study:
                input_rc = self.potential.rc
                input_mesh = np.copy(self.potential.pppm_mesh)
                input_alpha = self.potential.pppm_alpha_ewald

                self.timing_study = timing_study
                self.make_timing_plots()

                # Reset the original values.
                self.potential.rc = input_rc
                self.potential.pppm_mesh = np.copy(input_mesh)
                self.potential.pppm_alpha_ewald = input_alpha
                self.potential.setup(self.parameters)

            self.pppm_approximation()
            print('\nFigures can be found in {}'.format(self.pppm_plots_dir))

        if postprocessing:
            # POST- PROCESSING
            self.io.postprocess_info(self, write_to_file=True, observable='header')

            if hasattr(self, 'rdf'):
                self.rdf.setup(self.parameters)
                self.io.postprocess_info(self, write_to_file=True, observable='rdf')

            if hasattr(self, 'ssf'):
                self.ssf.setup(self.parameters)
                self.io.postprocess_info(self, write_to_file=True, observable='ssf')

            if hasattr(self, 'dsf'):
                self.dsf.setup(self.parameters)
                self.io.postprocess_info(self, write_to_file=True, observable='dsf')

            if hasattr(self, 'ccf'):
                self.ccf.setup(self.parameters)
                self.io.postprocess_info(self, write_to_file=True, observable='ccf')

            if hasattr(self, 'vm'):
                self.ccf.setup(self.parameters)
                self.io.postprocess_info(self, write_to_file=True, observable='vm')

    def make_timing_plots(self):
        """Estimate the best number of mesh points and cutoff radius."""

        from scipy.optimize import curve_fit

        print('\n\n{:=^70} \n'.format(' Timing Study '))

        max_cells = int(0.5 * self.parameters.box_lengths.min() / self.parameters.a_ws)
        if max_cells != self.pp_cells[-1]:
            self.pp_cells = np.arange(3, max_cells, dtype=int)

        pm_times = np.zeros(len(self.pm_meshes))
        pm_errs = np.zeros(len(self.pm_meshes))

        pp_times = np.zeros((len(self.pm_meshes), len(self.pp_cells)))
        pp_errs = np.zeros((len(self.pm_meshes), len(self.pp_cells)))

        pm_xlabels = []
        pp_xlabels = []

        self.force_error_map = np.zeros((len(self.pm_meshes), len(self.pp_cells)))

        # Average the PM time
        for i, m in enumerate(self.pm_meshes):

            self.potential.pppm_mesh = m * np.ones(3, dtype=int)
            self.potential.pppm_alpha_ewald = 0.3 * m / self.parameters.box_lengths.min()
            green_time = self.green_function_timer()
            pm_errs[i] = self.parameters.pppm_pm_err
            print('\n\nMesh = {} x {} x {} : '.format(*self.potential.pppm_mesh))
            print('alpha = {:.4f} / a_ws = {:.4e} '.format(self.potential.pppm_alpha_ewald * self.parameters.a_ws,
                                                           self.potential.pppm_alpha_ewald))
            print('PM Err = {:.6e}'.format(self.parameters.pppm_pm_err))

            self.io.preprocess_timing("GF", self.timer.time_division(green_time), 0)
            pm_xlabels.append("{}x{}x{}".format(*self.potential.pppm_mesh))

            # Calculate the PM acceleration timing 3x and average
            for it in range(3):
                self.timer.start()
                self.potential.update_pm(self.particles)
                pm_times[i] += self.timer.stop() / 3.0

            # For each number of PP cells, calculate the PM acceleration timing 3x and average
            for j, c in enumerate(self.pp_cells):
                self.potential.rc = self.parameters.box_lengths.min() / c
                kappa_over_alpha = - 0.25 * (self.kappa / self.potential.pppm_alpha_ewald) ** 2
                alpha_times_rcut = - (self.potential.pppm_alpha_ewald * self.potential.rc) ** 2
                # Update the Force error
                self.potential.pppm_pp_err = 2.0 * np.exp(kappa_over_alpha + alpha_times_rcut) / np.sqrt(
                    self.potential.rc)
                self.potential.pppm_pp_err *= np.sqrt(self.parameters.total_num_ptcls) * self.parameters.a_ws ** 2 \
                                              / np.sqrt(self.parameters.box_volume)

                pp_errs[i, j] = self.potential.pppm_pp_err
                self.force_error_map[i, j] = np.sqrt(self.potential.pppm_pp_err ** 2
                                                     + self.parameters.pppm_pm_err ** 2)

                if j == 0:
                    pp_xlabels.append("{:.2f}".format(self.potential.rc / self.parameters.a_ws))

                for it in range(3):
                    self.timer.start()
                    self.potential.update_linked_list(self.particles)
                    pp_times[i, j] += self.timer.stop() / 3.0

        # Get the time in seconds
        pp_times *= 1e-9
        pm_times *= 1e-9
        # Fit the PM times
        pm_popt, _ = curve_fit(
            lambda x, a, b: a + 5 * b * x ** 3 * np.log2(x**3),
            self.pm_meshes,
            pm_times)
        fit_str = r'Fit = $a_2 + 5 a_3 M^3 \log_2(M^3)$  [s]' + '\n' + r'$a_2 = ${:.4e}, $a_3 = ${:.4e} '.format(
            *pm_popt)
        print('\nPM Time ' + fit_str)

        # Fit the PP Times
        pp_popt, _ = curve_fit(
            lambda x, a, b: a + b / x ** 3,
            self.pp_cells,
            np.mean(pp_times, axis=0),
            p0=[np.mean(pp_times, axis =0)[0], self.parameters.total_num_ptcls],
            bounds = (0, [np.mean(pp_times, axis =0)[0], 1e9])
        )
        fit_pp_str = r'Fit = $a_0 + a_1 / N_c^3$  [s]' + '\n' + '$a_0 = ${:.4e},  $a_1 = ${:.4e}'.format(*pp_popt)
        print('\nPP Time ' + fit_pp_str)

        # Start the plot
        fig, (ax_pp, ax_pm) = plt.subplots(1, 2, sharey=True, figsize=(12, 7))
        ax_pm.plot(self.pm_meshes, pm_times, 'o', label='Measured')
        ax_pm.plot(
            self.pm_meshes,
            pm_popt[0] + 5 * pm_popt[1] * self.pm_meshes ** 3 * np.log2(self.pm_meshes**3),
            ls='--', label='Fit')
        ax_pm.set(title='PM calculation time and estimate', yscale = 'log', xlabel='Mesh size')
        ax_pm.set_xscale('log', base =2)
        ax_pm.legend(ncol=2)
        ax_pm.annotate(
            text=fit_str,
            xy=(self.pm_meshes[-1], pm_times[-1]),
            xytext=(self.pm_meshes[0], pm_times[-1]),
            bbox=dict(boxstyle="round4", fc="white", ec="k", lw=2)
        )

        # Scatter Plot the PP Times
        self.tot_time_map = np.zeros(pp_times.shape)
        for j, mesh_points in enumerate(self.pm_meshes):
            self.tot_time_map[j, :] = pm_times[j] + pp_times[j, :]
            ax_pp.plot(self.pp_cells, pp_times[j], 'o', label=r'@ Mesh {}$^3$'.format(mesh_points))

        # Plot the Fit PP times
        ax_pp.plot(self.pp_cells, pp_popt[0] + pp_popt[1]/ self.pp_cells ** 3, ls='--', label='Fit')
        ax_pp.legend(ncol=2)
        ax_pp.annotate(
            text=fit_pp_str,
            xy=(self.pp_cells[0], pp_times[0, 0]),
            xytext=(self.pp_cells[0], pp_times[-1, -1]),
            bbox=dict(boxstyle="round4", fc="white", ec="k", lw=2)
        )
        ax_pp.set(title='PP calculation time and estimate', yscale='log', ylabel='CPU Times [s]',
                  xlabel=r'$N_c $ = Cells')
        fig.tight_layout()
        fig.savefig(os.path.join(self.pppm_plots_dir, 'Times_' + self.io.job_id + '.png'))

        self.make_force_v_timing_plot()
        # self.lagrangian = np.empty((len(self.pm_meshes), len(self.pp_cells)))
        # self.tot_times = np.empty((len(self.pm_meshes), len(self.pp_cells)))
        # self.pp_times = np.copy(pp_times)
        # self.pm_times = np.copy(pm_times)
        # for i in range(len(self.pm_meshes)):
        #     self.tot_times[i, :] = pp_times[i] + pm_times[i]
        #     self.lagrangian[i, :] = self.force_error_map[i, :]
        #
        # best = np.unravel_index(self.lagrangian.argmin(), self.lagrangian.shape)
        # self.best_mesh = self.pm_meshes[best[0]]
        # self.best_cells = self.pp_cells[best[1]]

        # self.make_lagrangian_plot()
        #
        # # set the best parameter
        # self.potential.pppm_mesh = self.best_mesh * np.ones(3, dtype=int)
        # self.potential.rc = self.parameters.box_lengths.min() / self.best_cells
        # self.potential.pppm_alpha_ewald = 0.3 * self.best_mesh / self.parameters.box_lengths.min()
        # self.potential.pppm_setup(self.parameters)
        #
        # # print report
        # self.io.timing_study(self)
        # # time prediction
        # self.predicted_times = pp_times[best] + pm_times[best[0]]
        # # Print estimate of run times
        # self.io.time_stamp('Equilibration',
        #                    self.timer.time_division(self.predicted_times * self.integrator.equilibration_steps))
        # self.io.time_stamp('Production',
        #                    self.timer.time_division(self.predicted_times * self.integrator.production_steps))
        # self.io.time_stamp('Total Run',
        #                    self.timer.time_division(self.predicted_times * (self.integrator.equilibration_steps
        #                                                                     + self.integrator.production_steps)))

    def make_lagrangian_plot(self):

        c_mesh, m_mesh = np.meshgrid(self.pp_cells, self.pm_meshes)
        fig = plt.figure()
        ax = fig.add_subplot(111)  # projection='3d')
        # CS = ax.plot_surface(m_mesh, c_mesh, self.lagrangian, rstride=1, cstride=1, cmap='viridis', edgecolor='none')
        CS = ax.contourf(m_mesh,
                         c_mesh,
                         self.lagrangian,
                         norm=LogNorm(vmin=self.lagrangian.min(), vmax=self.lagrangian.max()))
        CS2 = ax.contour(CS, colors='w')
        ax.clabel(CS2, fmt='%1.0e', colors='w')
        fig.colorbar(CS)
        ax.scatter(self.best_mesh, self.best_cells, s=200, c='k')
        ax.set_xlabel('Mesh size')
        ax.set_ylabel(r'Cells = $L/r_c$')
        ax.set_title('2D Lagrangian')
        fig.savefig(os.path.join(self.io.preprocessing_dir, '2D_Lagrangian.png'))

    def make_force_v_timing_plot(self):

        # Plot the results
        fig_path = self.pppm_plots_dir
        c_mesh, m_mesh = np.meshgrid(self.pp_cells, self.pm_meshes)

        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(18, 9))
        if self.force_error_map.min() == 0.0:
            minv = 1e-120
        else:
            minv = self.force_error_map.min()

        maxt = self.force_error_map.max()
        nlvl = 12
        lvls = np.logspace(np.log10(minv), np.log10(maxt), nlvl)

        luxmap = mpl.cm.get_cmap('viridis', nlvl)
        luxnorm = mpl.colors.LogNorm(vmin=minv, vmax=maxt)
        CS = ax1.contourf(m_mesh,
                          c_mesh,
                          self.force_error_map,
                          levels=lvls,
                          cmap=luxmap,
                          norm=luxnorm)
        clb = fig.colorbar(mpl.cm.ScalarMappable(norm=luxnorm, cmap=luxmap), ax=ax1)
        clb.set_label(r'Force Error  [$Q^2/ a_{\rm ws}^2$]', rotation=270, va='bottom')
        CS2 = ax1.contour(CS, colors='w')
        ax1.clabel(CS2, fmt='%1.0e', colors='w')
        ax1.set_xlabel('Mesh size')
        ax1.set_ylabel(r'N_c = Cells')
        ax1.set_title('Force Error Map')

        # Timing Plot
        maxt = self.tot_time_map.max()
        mint = self.tot_time_map.min()
        # nlvl = 13
        lvls = np.logspace(np.log10(mint), np.log10(maxt), nlvl)
        luxmap = mpl.cm.get_cmap('viridis', nlvl)
        luxnorm = mpl.colors.LogNorm(vmin=minv, vmax=maxt)

        CS = ax2.contourf(m_mesh,
                          c_mesh,
                          self.tot_time_map,
                          levels=lvls, cmap=luxmap)
        CS2 = ax2.contour(CS, colors='w', levels=lvls)
        ax2.clabel(CS2, fmt='%.2e', colors='w')
        # fig.colorbar(, ax = ax2)
        clb = fig.colorbar(mpl.cm.ScalarMappable(norm=luxnorm, cmap=luxmap),
                           ax=ax2)
        clb.set_label('CPU Time [s]', rotation=270, va='bottom')

        ax2.set_xlabel('Mesh size')
        ax2.set_title('Timing Map')
        fig.savefig(os.path.join(fig_path, 'ForceErrorMap_v_Timing_' + self.io.job_id + '.png'))

    def time_acceleration(self):

        self.pp_acc_time = np.zeros(self.loops)
        for i in range(self.loops):
            self.timer.start()
            self.potential.update_linked_list(self.particles)
            self.pp_acc_time[i] = self.timer.stop()

        # Calculate the mean excluding the first value because that time include numba compilation time
        pp_mean_time = self.timer.time_division(np.mean(self.pp_acc_time[1:]))

        self.io.preprocess_timing("PP", pp_mean_time, self.loops)

        # PM acceleration
        if self.potential.pppm_on:
            self.pm_acc_time = np.zeros(self.loops)
            for i in range(self.loops):
                self.timer.start()
                self.potential.update_pm(self.particles)
                self.pm_acc_time[i] = self.timer.stop()
            pm_mean_time = self.timer.time_division(np.mean(self.pm_acc_time[1:]))
            self.io.preprocess_timing("PM", pm_mean_time, self.loops)

    def time_integrator_loop(self):
        """Run several loops of the equilibration and production phase to estimate the total time of the simulation."""
        if self.parameters.electrostatic_equilibration:
            # Save the original number of timesteps
            steps = np.array([self.integrator.equilibration_steps,
                              self.integrator.production_steps,
                              self.integrator.magnetization_steps])
            self.integrator.magnetization_steps = self.loops
        else:
            # Save the original number of timesteps
            steps = np.array([self.integrator.equilibration_steps,
                              self.integrator.production_steps])

        # Update the equilibration and production timesteps for estimation
        self.integrator.production_steps = self.loops
        self.integrator.equilibration_steps = self.loops

        if self.io.verbose:
            print('\nRunning {} equilibration and production steps to estimate simulation times\n'.format(self.loops))

        # Run few equilibration steps to estimate the equilibration time
        self.timer.start()
        self.integrator.equilibrate(0, self.particles, self.io)
        self.eq_mean_time = self.timer.stop() / self.loops
        # Print the average equilibration & production times
        self.io.preprocess_timing("Equilibration", self.timer.time_division(self.eq_mean_time), self.loops)

        if self.integrator.electrostatic_equilibration:
            self.timer.start()
            self.integrator.magnetize(0, self.particles, self.io)
            self.mag_mean_time = self.timer.stop() / self.loops
            # Print the average equilibration & production times
            self.io.preprocess_timing("Magnetization", self.timer.time_division(self.mag_mean_time), self.loops)

        # Run few production steps to estimate the equilibration time
        self.timer.start()
        self.integrator.produce(0, self.particles, self.io)
        self.prod_mean_time = self.timer.stop() / self.loops
        self.io.preprocess_timing("Production", self.timer.time_division(self.prod_mean_time), self.loops)

        # Restore the original number of timesteps and print an estimate of run times
        self.integrator.equilibration_steps = steps[0]
        self.integrator.production_steps = steps[1]

        # Print the estimate for the full run
        eq_prediction = self.eq_mean_time * steps[0]
        self.io.time_stamp('Equilibration', self.timer.time_division(eq_prediction))

        if self.integrator.electrostatic_equilibration:
            self.integrator.magnetization_steps = steps[2]
            mag_prediction = self.mag_mean_time * steps[2]
            self.io.time_stamp('Magnetization', self.timer.time_division(mag_prediction))
            eq_prediction += mag_prediction

        prod_prediction = self.prod_mean_time * steps[1]
        self.io.time_stamp('Production', self.timer.time_division(prod_prediction))

        tot_time = eq_prediction + prod_prediction
        self.io.time_stamp('Total Run', self.timer.time_division(tot_time))

    def pppm_approximation(self):
        """Calculate the Force error for a PPPM simulation using analytical approximations.
        Plot the force error in the parameter space."""

        # Calculate Force error from analytic approximation given in Dharuman et al. J Chem Phys 2017
        total_force_error, pp_force_error, pm_force_error, rcuts, alphas = self.analytical_approx_pppm()

        chosen_alpha = self.potential.pppm_alpha_ewald * self.parameters.a_ws
        chosen_rcut = self.potential.rc / self.parameters.a_ws

        # mesh_dir = os.path.join(self.pppm_plots_dir, 'Mesh_{}'.format(self.potential.pppm_mesh[0]))
        # if not os.path.exists(mesh_dir):
        #     os.mkdir(mesh_dir)
        #
        # cell_num = int(self.parameters.box_lengths.min() / self.potential.rc)
        # cell_dir = os.path.join(mesh_dir, 'Cells_{}'.format(cell_num))
        # if not os.path.exists(cell_dir):
        #     os.mkdir(cell_dir)
        #
        # self.pppm_plots_dir = cell_dir

        # Color Map
        self.make_color_map(rcuts, alphas, chosen_alpha, chosen_rcut, total_force_error)

        # Line Plot
        self.make_line_plot(rcuts, alphas, chosen_alpha, chosen_rcut, total_force_error)

    def make_line_plot(self, rcuts, alphas, chosen_alpha, chosen_rcut, total_force_error):
        """
        Plot selected values of the total force error approximation.

        Parameters
        ----------
        rcuts: numpy.ndarray
            Cut off distances.

        alphas: numpy.ndarray
            Ewald parameters.

        chosen_alpha: float
            Chosen Ewald parameter.

        chosen_rcut: float
            Chosen cut off radius.

        total_force_error: numpy.ndarray
            Force error matrix.

        """
        # Plot the results
        fig_path = self.pppm_plots_dir

        fig, ax = plt.subplots(1, 2, constrained_layout=True, figsize=(12, 7))
        ax[0].plot(rcuts, total_force_error[30, :], ls= (0, (5, 10)), label=r'$\alpha a_{ws} = ' + '{:2.2f}$'.format(alphas[30]))
        ax[0].plot(rcuts, total_force_error[40, :], ls= 'dashed', label=r'$\alpha a_{ws} = ' + '{:2.2f}$'.format(alphas[40]))
        ax[0].plot(rcuts, total_force_error[50, :], ls = 'solid', label=r'$\alpha a_{ws} = ' + '{:2.2f}$'.format(alphas[50]))
        ax[0].plot(rcuts, total_force_error[60, :], ls = 'dashdot',label=r'$\alpha a_{ws} = ' + '{:2.2f}$'.format(alphas[60]))
        ax[0].plot(rcuts, total_force_error[70, :], ls = (0, (3, 10, 1, 10)),label=r'$\alpha a_{ws} = ' + '{:2.2f}$'.format(alphas[70]))
        ax[0].set_ylabel(r'$\Delta F^{approx}_{tot}$')
        ax[0].set_xlabel(r'$r_c/a_{ws}$')
        ax[0].set_yscale('log')
        ax[0].axvline(chosen_rcut, ls='--', c='k')
        ax[0].axhline(self.parameters.force_error, ls='--', c='k')
        if rcuts[-1] * self.parameters.a_ws > 0.5 * self.parameters.box_lengths.min():
            ax[0].axvline(0.5 * self.parameters.box_lengths.min() / self.parameters.a_ws, c='r', label=r'$L/2$')
        ax[0].grid(True, alpha=0.3)
        ax[0].legend(loc='best')

        ax[1].plot(alphas, total_force_error[:, 30], ls = (0, (5, 10)),label=r'$r_c = {:2.2f}'.format(rcuts[30]) + ' a_{ws}$')
        ax[1].plot(alphas, total_force_error[:, 40], ls = 'dashed',label=r'$r_c = {:2.2f}'.format(rcuts[40]) + ' a_{ws}$')
        ax[1].plot(alphas, total_force_error[:, 50], ls = 'solid',label=r'$r_c = {:2.2f}'.format(rcuts[50]) + ' a_{ws}$')
        ax[1].plot(alphas, total_force_error[:, 60], ls = 'dashdot',label=r'$r_c = {:2.2f}'.format(rcuts[60]) + ' a_{ws}$')
        ax[1].plot(alphas, total_force_error[:, 70], ls = (0, (3, 10, 1, 10)), label=r'$r_c = {:2.2f}'.format(rcuts[70]) + ' a_{ws}$')
        ax[1].set_xlabel(r'$\alpha \; a_{ws}$')
        ax[1].set_yscale('log')
        ax[1].axhline(self.parameters.force_error, ls='--', c='k')
        ax[1].axvline(chosen_alpha, ls='--', c='k')
        ax[1].grid(True, alpha=0.3)
        ax[1].legend(loc='best')
        fig.suptitle(
            r'Parameters  $N = {}, \quad M = {}, \quad p = {}, \quad \kappa = {:.2f}$'.format(
                self.parameters.total_num_ptcls,
                self.potential.pppm_mesh[0],
                self.potential.pppm_cao,
                self.kappa * self.parameters.a_ws))
        fig.savefig(os.path.join(fig_path, 'LinePlot_ForceError_' + self.io.job_id + '.png'))

    def make_color_map(self, rcuts, alphas, chosen_alpha, chosen_rcut, total_force_error):
        """
        Plot a color map of the total force error approximation.

        Parameters
        ----------
        rcuts: numpy.ndarray
            Cut off distances.

        alphas: numpy.ndarray
            Ewald parameters.

        chosen_alpha: float
            Chosen Ewald parameter.

        chosen_rcut: float
            Chosen cut off radius.

        total_force_error: numpy.ndarray
            Force error matrix.
        """
        # Plot the results
        fig_path = self.pppm_plots_dir

        r_mesh, a_mesh = np.meshgrid(rcuts, alphas)
        fig, ax = plt.subplots(1, 1, figsize=(10, 7))
        if total_force_error.min() == 0.0:
            minv = 1e-120
        else:
            minv = total_force_error.min()
        total_force_error[total_force_error == 0.0] = minv
        CS = ax.contourf(a_mesh, r_mesh, total_force_error, norm=LogNorm(vmin=minv, vmax=total_force_error.max()))
        CS2 = ax.contour(CS, colors='w')
        ax.clabel(CS2, fmt='%1.0e', colors='w')
        ax.scatter(chosen_alpha, chosen_rcut, s=200, c='k')
        if rcuts[-1] * self.parameters.a_ws > 0.5 * self.parameters.box_lengths.min():
            ax.axhline(0.5 * self.parameters.box_lengths.min() / self.parameters.a_ws, c='r', label=r'$L/2$')
        # ax.tick_parameters(labelsize=fsz)
        ax.set_xlabel(r'$\alpha \;a_{ws}$')
        ax.set_ylabel(r'$r_c/a_{ws}$')
        ax.set_title(r'Parameters  $N = {}, \quad M = {}, \quad p = {}, \quad \kappa = {:.2f}$'.format(
            self.parameters.total_num_ptcls,
            self.potential.pppm_mesh[0],
            self.potential.pppm_cao,
            self.kappa * self.parameters.a_ws))
        clb = fig.colorbar(CS)
        clb.set_label(r'$\Delta F^{approx}_{tot}(r_c,\alpha)$', va='bottom', rotation=270)
        fig.tight_layout()
        fig.savefig(os.path.join(fig_path, 'ClrMap_ForceError_' + self.io.job_id + '.png'))

    def analytical_approx_pp(self):
        """Calculate PP force error."""

        r_min = self.potential.rc * 0.5
        r_max = self.potential.rc * 1.5

        rcuts = np.linspace(r_min, r_max, 101) / self.parameters.a_ws

        # Calculate the analytic PP error and the total force error
        pp_force_error = np.sqrt(2.0 * np.pi * self.kappa) * np.exp(- rcuts * self.kappa)
        pp_force_error *= np.sqrt(self.parameters.total_num_ptcls *
                                  self.parameters.a_ws ** 3 / self.parameters.box_volume)

        return pp_force_error, rcuts

    def analytical_approx_pppm(self):
        """Calculate the total force error as given in Dharuman et al. J Chem Phys 146 024112 (2017)."""

        p = self.potential.pppm_cao
        L = self.parameters.box_lengths[0] / self.parameters.a_ws
        h = L / self.potential.pppm_mesh[0]

        a_min = self.potential.pppm_alpha_ewald * 0.5
        a_max = self.potential.pppm_alpha_ewald * 1.5

        r_min = self.potential.rc * 0.5
        r_max = self.potential.rc * 1.5

        alphas = self.parameters.a_ws * np.linspace(a_min, a_max, 101)
        rcuts = np.linspace(r_min, r_max, 101) / self.parameters.a_ws

        pm_force_error = np.zeros(len(alphas))
        pp_force_error = np.zeros((len(alphas), len(rcuts)))
        total_force_error = np.zeros((len(alphas), len(rcuts)))

        # Coefficient from Deserno and Holm J Chem Phys 109 7694 (1998)
        if p == 1:
            Cmp = np.array([2 / 3])
        elif p == 2:
            Cmp = np.array([2 / 45, 8 / 189])
        elif p == 3:
            Cmp = np.array([4 / 495, 2 / 225, 8 / 1485])
        elif p == 4:
            Cmp = np.array([2 / 4725, 16 / 10395, 5528 / 3869775, 32 / 42525])
        elif p == 5:
            Cmp = np.array([4 / 93555, 2764 / 11609325, 8 / 25515, 7234 / 32531625, 350936 / 3206852775])
        elif p == 6:
            Cmp = np.array([2764 / 638512875, 16 / 467775, 7234 / 119282625, 1403744 / 25196700375,
                            1396888 / 40521009375, 2485856 / 152506344375])
        elif p == 7:
            Cmp = np.array([8 / 18243225, 7234 / 1550674125, 701872 / 65511420975, 2793776 / 225759909375,
                            1242928 / 132172165125, 1890912728 / 352985880121875, 21053792 / 8533724574375])

        kappa = self.kappa * self.parameters.a_ws

        for ia, alpha in enumerate(alphas):
            somma = 0.0
            for m in np.arange(p):
                expp = 2 * (m + p)
                somma += Cmp[m] * (2 / (1 + expp)) * betamp(m, p, alpha, kappa) * (h / 2.) ** expp
            # eq.(36) in Dharuman J Chem Phys 146 024112 (2017)
            pm_force_error[ia] = np.sqrt(3.0 * somma) / (2.0 * np.pi)
        # eq.(35)
        pm_force_error *= np.sqrt(self.parameters.total_num_ptcls *
                                  self.parameters.a_ws ** 3 / self.parameters.box_volume)
        # Calculate the analytic PP error and the total force error
        if self.potential.type == "QSP":
            for (ir, rc) in enumerate(rcuts):
                pp_force_error[:, ir] = np.sqrt(2.0 * np.pi * kappa) * np.exp(- rc * kappa)
                pp_force_error[:, ir] *= np.sqrt(self.parameters.total_num_ptcls
                                                 * self.parameters.a_ws ** 3 / self.parameters.box_volume)
                for (ia, alfa) in enumerate(alphas):
                    # eq.(42) from Dharuman J Chem Phys 146 024112 (2017)
                    total_force_error[ia, ir] = np.sqrt(pm_force_error[ia] ** 2 + pp_force_error[ia, ir] ** 2)
        else:
            for (ir, rc) in enumerate(rcuts):
                for (ia, alfa) in enumerate(alphas):
                    # eq.(30) from Dharuman J Chem Phys 146 024112 (2017)
                    pp_force_error[ia, ir] = 2.0 * np.exp(-(0.5 * kappa / alfa) ** 2
                                                          - alfa ** 2 * rc ** 2) / np.sqrt(rc)
                    pp_force_error[ia, ir] *= np.sqrt(self.parameters.total_num_ptcls *
                                                      self.parameters.a_ws ** 3 / self.parameters.box_volume)
                    # eq.(42) from Dharuman J Chem Phys 146 024112 (2017)
                    total_force_error[ia, ir] = np.sqrt(pm_force_error[ia] ** 2 + pp_force_error[ia, ir] ** 2)

        return total_force_error, pp_force_error, pm_force_error, rcuts, alphas

    @staticmethod
    def make_fit_plot(pp_xdata, pm_xdata, pp_times, pm_times, pp_opt, pm_opt, pp_xlabels, pm_xlabels, fig_path):
        """
        Make a dual plot of the fitted functions.
        """
        fig, ax = plt.subplots(1, 2, figsize=(12, 7))
        ax[0].plot(pm_xdata, pm_times.mean(axis=-1), 'o', label='Measured times')
        # ax[0].plot(pm_xdata, quadratic(pm_xdata, *pm_opt), '--r', label="Fit $f(x) = a + b x + c x^2$")
        ax[1].plot(pp_xdata, pp_times.mean(axis=-1), 'o', label='Measured times')
        # ax[1].plot(pp_xdata, linear(pp_xdata, *pp_opt), '--r', label="Fit $f(x) = a x$")

        ax[0].set_xscale('log')
        ax[0].set_yscale('log')

        ax[1].set_xscale('log')
        ax[1].set_yscale('log')

        ax[0].legend()
        ax[1].legend()

        ax[0].set_xticks(pm_xdata)
        ax[0].set_xticklabels(pm_xlabels)
        # Rotate the tick labels and set their alignment.
        plt.setp(ax[0].get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

        ax[1].set_xticks(pp_xdata[0:-1:3])
        ax[1].set_xticklabels(pp_xlabels)
        # Rotate the tick labels and set their alignment.
        plt.setp(ax[1].get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

        ax[0].set_title("PM calculation")
        ax[1].set_title("PP calculation")

        ax[0].set_xlabel('Mesh sizes')
        ax[1].set_xlabel(r'$r_c / a_{ws}$')
        fig.tight_layout()
        fig.savefig(os.path.join(fig_path, 'Timing_Fit.png'))


class Simulation(Process):
    """
    Sarkas simulation wrapper. This class manages the entire simulation and its small moving parts.

    Parameters
    ----------
    input_file : str
        Path to the YAML input file.

    """

    def __init__(self, input_file: str = None):
        self.__name__ = 'simulation'
        super().__init__(input_file)

    def equilibrate(self) -> None:
        """
        Run the time integrator with the thermostat to evolve the system to its thermodynamics equilibrium state.
        """
        if self.parameters.verbose:
            print("\n------------- Equilibration -------------")
        # Check if this is restart
        if self.parameters.load_method in ["equilibration_restart", "eq_restart"]:
            it_start = self.parameters.restart_step
        else:
            it_start = 0
            self.io.dump('equilibration', self.particles, 0)

        # Start timer, equilibrate, and print run time.
        self.timer.start()
        self.integrator.equilibrate(it_start, self.particles, self.io)
        time_eq = self.timer.stop()
        self.io.time_stamp("Equilibration", self.timer.time_division(time_eq))

        # Check for magnetization phase
        if self.integrator.electrostatic_equilibration:
            if self.parameters.verbose:
                print('\n------------- Magnetization -------------')

            if self.parameters.load_method in ["magnetization_restart", "mag_restart"]:
                it_start = self.parameters.restart_step
            else:
                it_start = 0
                self.io.dump('magnetization', self.particles, it_start)

            # Start timer, magnetize, and print run time.
            self.timer.start()
            self.integrator.magnetize(it_start, self.particles, self.io)
            time_eq = self.timer.stop()
            self.io.time_stamp("Magnetization", self.timer.time_division(time_eq))

    def evolve(self) -> None:
        """
        Run the time integrator to evolve the system for the duration of the production phase.
        """

        # Check for simulation restart.
        if self.parameters.load_method in ["prod_restart", "production_restart"]:
            it_start = self.parameters.restart_step
        else:
            it_start = 0
            # Restart the pbc counter.
            self.particles.pbc_cntr.fill(0)
            self.io.dump('production', self.particles, 0)

        if self.parameters.verbose:
            print("\n------------- Production -------------")

        # Update measurement flag for rdf.
        self.potential.measure = True

        # Start timer, produce data, and print run time.
        self.timer.start()
        self.integrator.produce(it_start, self.particles, self.io)
        time_end = self.timer.stop()
        self.io.time_stamp("Production", self.timer.time_division(time_end))

    def run(self) -> None:
        """Run the simulation."""
        time0 = self.timer.current()

        if not self.parameters.load_method in ['prod_restart', 'production_restart']:
            self.equilibrate()

        self.evolve()
        time_tot = self.timer.current()
        self.io.time_stamp("Total", self.timer.time_division(time_tot - time0))

    # def setup(self, read_yaml: bool = False, other_inputs=None):
    #     """Setup simulations' parameters and io subclasses.
    #
    #     Parameters
    #     ----------
    #     read_yaml: bool
    #         Flag for reading YAML input file. Default = False.
    #
    #     other_inputs: dict (optional)
    #         Dictionary with additional simulations options.
    #
    #     """
    #     if read_yaml:
    #         self.common_parser()
    #
    #     if other_inputs:
    #         if not isinstance(other_inputs, dict):
    #             raise TypeError("Wrong input type. other_inputs should be a nested dictionary")
    #
    #         for class_name, class_attr in other_inputs.items():
    #             if not class_name == 'Particles':
    #                 self.__dict__[class_name.lower()].from_dict(class_attr)
    #             else:
    #                 for sp, species in enumerate(other_inputs["Particles"]):
    #                     spec = Species(species["Species"])
    #                     self.species[sp].__dict__.update(spec.__dict__)
    #     # initialize the directories and filenames
    #     self.io.setup()
    #     # Copy relevant subsclasses attributes into parameters class. This is needed for post-processing.
    #
    #     # Update parameters' dictionary with filenames and directories
    #     self.parameters.from_dict(self.io.__dict__)
    #     # save some general info
    #     self.parameters.potential_type = self.potential.type
    #     self.parameters.cutoff_radius = self.potential.rc
    #
    #     # Copy some integrator parameters if not already defined
    #     if not hasattr(self.parameters, 'dt'):
    #         self.parameters.dt = self.integrator.dt
    #     if not hasattr(self.parameters, 'equilibration_steps'):
    #         self.parameters.equilibration_steps = self.integrator.equilibration_steps
    #     if not hasattr(self.parameters, 'eq_dump_step'):
    #         self.parameters.eq_dump_step = self.integrator.eq_dump_step
    #     if not hasattr(self.parameters, 'production_steps'):
    #         self.parameters.production_steps = self.integrator.production_steps
    #     if not hasattr(self.parameters, 'prod_dump_step'):
    #         self.parameters.prod_dump_step = self.integrator.prod_dump_step
    #
    #     if self.integrator.electrostatic_equilibration:
    #         self.parameters.electrostatic_equilibration = True
    #         if not hasattr(self.parameters, 'mag_dump_step'):
    #             self.parameters.mag_dump_step = self.integrator.mag_dump_step
    #         if not hasattr(self.parameters, 'magnetization_steps'):
    #             self.parameters.magnetization_steps = self.integrator.magnetization_steps
    #
    #     self.parameters.setup(self.species)
    #
    #     self.io.setup_checkpoint(self.parameters, self.species)


@njit
def Gk(x, alpha, kappa):
    """
    Green's function of Coulomb/Yukawa potential.
    """
    return 4.0 * np.pi * np.exp(-(x ** 2 + kappa ** 2) / (2 * alpha) ** 2) / (kappa ** 2 + x ** 2)


@njit
def betamp(m, p, alpha, kappa):
    """
    Calculate :math:`\beta(m)` of eq.(37) in Dharuman et al. J Chem Phys 146 024112 (2017)
    """
    xa = np.linspace(0.0001, 500, 5000)
    return np.trapz(Gk(xa, alpha, kappa) * Gk(xa, alpha, kappa) * xa ** (2 * (m + p + 2)), x=xa)


@njit
def analytical_approx_pppm_single(kappa, rc, p, h, alpha):
    """
    Calculate the total force error for a given value of ``rc`` and ``alpha``. See similar function above.
    """
    # Coefficient from Deserno and Holm J Chem Phys 109 7694 (1998)
    if p == 1:
        Cmp = np.array([2 / 3])
    elif p == 2:
        Cmp = np.array([2 / 45, 8 / 189])
    elif p == 3:
        Cmp = np.array([4 / 495, 2 / 225, 8 / 1485])
    elif p == 4:
        Cmp = np.array([2 / 4725, 16 / 10395, 5528 / 3869775, 32 / 42525])
    elif p == 5:
        Cmp = np.array([4 / 93555, 2764 / 11609325, 8 / 25515, 7234 / 32531625, 350936 / 3206852775])
    elif p == 6:
        Cmp = np.array([2764 / 638512875, 16 / 467775, 7234 / 119282625, 1403744 / 25196700375,
                        1396888 / 40521009375, 2485856 / 152506344375])
    elif p == 7:
        Cmp = np.array([8 / 18243225, 7234 / 1550674125, 701872 / 65511420975, 2793776 / 225759909375,
                        1242928 / 132172165125, 1890912728 / 352985880121875, 21053792 / 8533724574375])

    somma = 0.0
    for m in np.arange(p):
        expp = 2 * (m + p)
        somma += Cmp[m] * (2 / (1 + expp)) * betamp(m, p, alpha, kappa) * (h / 2.) ** expp
    # eq.(36) in Dharuman J Chem Phys 146 024112 (2017)
    pm_force_error = np.sqrt(3.0 * somma) / (2.0 * np.pi)

    # eq.(30) from Dharuman J Chem Phys 146 024112 (2017)
    pp_force_error = 2.0 * np.exp(-(0.5 * kappa / alpha) ** 2 - alpha ** 2 * rc ** 2) / np.sqrt(rc)
    # eq.(42) from Dharuman J Chem Phys 146 024112 (2017)
    Tot_DeltaF = np.sqrt(pm_force_error ** 2 + pp_force_error ** 2)

    return Tot_DeltaF, pp_force_error, pm_force_error


def pm_time_model(x, a, b):
    """
    Quadratic function for fitting.

    Parameters
    ----------
    x : array
        Values at which to calculate the function.

    a: float
        Intercept.

    b: float
        Coefficient of linear term.

    c: float
        Coefficient of quadratic term.

    Returns
    -------
    quadratic formula
    """
    return a * x + b * x * np.log(x)



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# Sarkas Modules
import sarkas.tools.observables as obs


class TransportCoefficient:
    """
    Transport Coefficients class.
    """

    def __init__(self):
        pass

    def __repr__(self):
        sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
        disp = 'Transport( \n'
        for key, value in sortedDict.items():
            disp += "\t{} : {}\n".format(key, value)
        disp += ')'
        return disp

    @staticmethod
    def electrical_conductivity(params,
                  phase: str = 'production',
                  compute_acf: bool = True,
                  no_slices: int = 1,
                  plot: bool = True,
                  show: bool = False,
                  figname: str = None,
                  **kwargs):
        """
        Calculate electrical conductivity from current auto-correlation function.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters.

        phase : str
            Phase to analyze. Default = 'production'.

        show : bool
            Flag for prompting plot to screen.

        Returns
        -------
        coefficient : pandas.DataFrame
            Pandas dataframe containing the value of the transport coefficient as a function of integration time.

        """

        print('\n\n{:=^70} \n'.format(' Electrical Conductivity '))
        coefficient = pd.DataFrame()

        if compute_acf:
            jc_acf = obs.ElectricCurrent()
            jc_acf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            jc_acf.compute()

        else:
            jc_acf = obs.ElectricCurrent()
            jc_acf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            jc_acf.parse()

        # Print some info
        print('Data saved in: \n', os.path.join(jc_acf.saving_dir, 'ElectricalConductivity_' + jc_acf.job_id + '.h5'))

        print('\nNo. of slices = {}'.format(jc_acf.no_slices))
        print('No. dumps per slice = {}'.format( int(jc_acf.slice_steps/jc_acf.dump_step) ))

        print('Time interval of autocorrelation function = {:.4e} [s] ~ {} w_p T'.format(
            jc_acf.dt * jc_acf.slice_steps,
            int(jc_acf.dt * jc_acf.slice_steps * jc_acf.total_plasma_frequency)))

        no_int = jc_acf.slice_steps

        # to_numpy creates a 2d-array, hence the [:,0]
        time = jc_acf.dataframe[("Time")].to_numpy()[:, 0]

        coefficient["Time"] = time
        jc_str = "Electric Current ACF"
        sigma_str = "Electrical Conductivity"
        for isl in range(jc_acf.no_slices):
            sigma_ij = np.zeros(jc_acf.slice_steps)

            integrand = np.array(jc_acf.dataframe[(jc_str, "Total", "slice {}".format(isl))])

            for it in range(1, no_int):
                sigma_ij[it] = np.trapz(integrand[:it]/integrand[0], x=time[:it])

                coefficient[sigma_str + "_slice {}".format(isl)] = sigma_ij[:]

        col_str = [sigma_str + "_slice {}".format(isl) for isl in range(jc_acf.no_slices)]
        coefficient[sigma_str + "_Mean"] = coefficient[col_str].mean(axis=1)
        coefficient[sigma_str + "_Std"] = coefficient[col_str].std(axis=1)

        coefficient.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in coefficient.columns])
        coefficient.to_hdf(
            os.path.join(jc_acf.saving_dir, 'ElectricalConductivity_' + jc_acf.job_id + '.h5'),
            mode='w',
            key='conductivity',
            index=False)

        if plot or figname:
            # Make the plot
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 7))
            ax3 = ax1.twiny()
            ax4 = ax2.twiny()
            # extra space for the second axis at the bottom
            # fig.subplots_adjust(bottom=0.2)

            acf_avg = jc_acf.dataframe[(jc_str, "Total", "Mean")]
            acf_std = jc_acf.dataframe[(jc_str, "Total", "Std")]

            d_avg = coefficient[(sigma_str, "Mean")]
            d_std = coefficient[(sigma_str, "Std")]

            # Calculate axis multipliers and labels
            xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, d_avg, "Time", "Conductivity", jc_acf.units)

            # ACF
            ax1.plot(xmul * time, acf_avg / acf_avg.iloc[0], label=r'Current $J$')
            ax1.fill_between(
                xmul * time,
                (acf_avg + acf_std) / (acf_avg.iloc[0] + acf_std.iloc[0]),
                (acf_avg - acf_std) / (acf_avg.iloc[0] - acf_std.iloc[0]), alpha=0.2)

            # Coefficient
            ax2.plot(xmul * time, ymul * d_avg, label=r'$\sigma$')
            ax2.fill_between(xmul * time, ymul * (d_avg + d_std), ymul * (d_avg - d_std), alpha=0.2)

            xlims = (xmul * time[1], xmul * time[-1] * 1.5)

            ax1.set(xlim=xlims, xscale='log', ylabel=r'Electric Current ACF', xlabel=r"Time difference" + xlbl)
            ax2.set(xlim=xlims, xscale='log', ylabel=r'Conductivity' + ylbl, xlabel=r"$\tau$" + xlbl)

            ax1.legend(loc='best')
            ax2.legend(loc='best')
            # Finish the index axes
            for axi in [ax3, ax4]:
                axi.grid(alpha=0.1)
                axi.set(xlim=(1, jc_acf.slice_steps * 1.5), xscale='log', xlabel='Index')

            fig.tight_layout()
            if figname:
                fig.savefig(os.path.join(jc_acf.saving_dir, figname))
            else:
                fig.savefig(os.path.join(jc_acf.saving_dir, 'Plot_ElectricConductivity_' + jc_acf.job_id + '.png'))

            if show:
                fig.show()

        return coefficient

    @staticmethod
    def diffusion(params,
                  phase: str = 'production',
                  compute_acf: bool = True,
                  no_slices: int = 1,
                  plot: bool = True,
                  show: bool = False,
                  figname: str = None,
                  **kwargs):
        """
        Calculate the self-diffusion coefficient from the velocity auto-correlation function.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters.

        phase : str, optional
            Phase to analyze. Default = 'production'.

        compute_acf : bool, optional
            Flag for recalculating the ACF. Default = True.
            If False it will read in the data from the dataframe.

        no_slices : int, optional
            Number of slices of the simulation. Default = 1.

        plot : bool, optional
            Flag to plot transport coefficient with corresponding autocorrelation function. Default = True.

        show : bool, optional
            Flag for prompting plot to screen.

        figname : str, optional
            Name with which to save the file. It automatically saves it in the correct directory.

        **kwargs:
            Arguments to pass :meth:`sarkas.tools.observables.VelocityAutoCorrelationFunction`

        Returns
        -------
        coefficient : pandas.DataFrame
            Pandas dataframe containing the value of the transport coefficient as a function of integration time.

        """
        print('\n\n{:=^70} \n'.format(' Diffusion Coefficient '))
        coefficient = pd.DataFrame()
        if compute_acf:
            vacf = obs.VelocityAutoCorrelationFunction()
            vacf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            vacf.compute()

        else:
            vacf = obs.VelocityAutoCorrelationFunction()
            vacf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            vacf.parse()

        print('Data saved in: \n', os.path.join(vacf.saving_dir, 'Diffusion_' + vacf.job_id + '.h5'))

        print('\nNo. of slices = {}'.format(vacf.no_slices))
        print('No. dumps per slice = {}'.format(int(vacf.slice_steps/vacf.dump_step)))

        print('Time interval of autocorrelation function = {:.4e} [s] ~ {} w_p T'.format(
            vacf.dt * vacf.slice_steps,
            int(vacf.dt * vacf.slice_steps * vacf.total_plasma_frequency)))

        vacf.compute()
        time = vacf.dataframe["Time"].to_numpy()[:,0]
        coefficient["Time"] = time
        vacf_str = 'VACF'
        const = 1.0 / 3.0
        if not params.magnetized:
            # Loop over time slices
            for isl in range(vacf.no_slices):
                # Initialize the temporary diffusion container
                D = np.zeros((params.num_species, vacf.slice_steps))
                # Iterate over the number of species
                for i, sp in enumerate(params.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    # Grab vacf data of each slice
                    integrand = np.array(vacf.dataframe[(sp_vacf_str, "Total","slice {}".format(isl) )])
                    # Integrate each timestep
                    for it in range(1, len(time)):
                        D[i, it] = const * np.trapz(integrand[:it], x=time[:it])

                    coefficient["{} Diffusion_slice {}".format(sp, isl)] = D[i, :]

            # Average and std of each diffusion coefficient.
            for isp, sp in enumerate(params.species_names):
                col_str = ["{} Diffusion_slice {}".format(sp, isl) for isl in range(vacf.no_slices)]

                coefficient["{} Diffusion_Mean".format(sp)] = coefficient[col_str].mean(axis=1)
                coefficient["{} Diffusion_Std".format(sp)] = coefficient[col_str].std(axis=1)

        else:
            # Loop over time slices
            for isl in range(vacf.no_slices):
                # Initialize the temporary diffusion container
                D = np.zeros((params.num_species, 2, len(time)))
                # Iterate over the number of species
                for i, sp in enumerate(params.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    integrand_par = np.array(vacf.dataframe[(sp_vacf_str, 'Z', "slice {}".format(isl))])
                    integrand_perp = np.array(vacf.dataframe[(sp_vacf_str, 'X', "slice {}".format(isl))]) + \
                                     np.array(vacf.dataframe[(sp_vacf_str, 'Y', "slice {}".format(isl))])

                    for it in range(1, len(time)):
                        D[i, 0, it] = np.trapz(integrand_par[:it], x=time[:it])
                        D[i, 1, it] = 0.5 * np.trapz(integrand_perp[:it], x=time[:it])

                    coefficient["{} Parallel Diffusion_slice {}".format(sp, isl)] = D[i, 0, :]
                    coefficient["{} Perpendicular Diffusion_slice {}".format(sp, isl)] = D[i, 1, :]

            # Add the average and std of perp and par VACF to its dataframe
            for isp, sp in enumerate(params.species_names):
                par_col_str = ["{} Z Velocity ACF slice {}".format(sp, isl) for isl in range(vacf.no_slices)]

                vacf.dataframe["{} Parallel Velocity ACF avg".format(sp)] = vacf.dataframe[par_col_str].mean(axis=1)
                vacf.dataframe["{} Parallel Velocity ACF std".format(sp)] = vacf.dataframe[par_col_str].std(axis=1)

                x_col_str = ["{} X Velocity ACF slice {}".format(sp, isl) for isl in range(vacf.no_slices)]
                y_col_str = ["{} Y Velocity ACF slice {}".format(sp, isl) for isl in range(vacf.no_slices)]

                perp_vacf = 0.5 * (np.array(vacf.dataframe[x_col_str]) + np.array(vacf.dataframe[y_col_str]))
                vacf.dataframe["{} Perpendicular Velocity ACF avg".format(sp)] = perp_vacf.mean(axis=1)
                vacf.dataframe["{} Perpendicular Velocity ACF std".format(sp)] = perp_vacf.std(axis=1)

                # Average and std of each diffusion coefficient.
                par_col_str = ["{} Parallel Diffusion slice {}".format(sp, isl) for isl in range(vacf.no_slices)]
                perp_col_str = ["{} Perpendicular Diffusion slice {}".format(sp, isl) for isl in range(vacf.no_slices)]

                coefficient["{} Parallel Diffusion avg".format(sp)] = coefficient[par_col_str].mean(axis=1)
                coefficient["{} Parallel Diffusion std".format(sp)] = coefficient[par_col_str].std(axis=1)

                coefficient["{} Perpendicular Diffusion avg".format(sp)] = coefficient[perp_col_str].mean(axis=1)
                coefficient["{} Perpendicular Diffusion std".format(sp)] = coefficient[perp_col_str].std(axis=1)

            # Save the updated dataframe
            vacf.dataframe.to_csv(vacf.filename_csv, index=False, encoding='utf-8')

        # Endif magnetized.
        coefficient.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in coefficient.columns])
        # Save the coefficient's data
        coefficient.to_hdf(
            os.path.join(vacf.saving_dir, 'Diffusion_' + vacf.job_id + '.h5'),
            mode='w',
            key='diffusion')

        if plot or figname:
            # Make the plot
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 7))
            # index axes
            ax3 = ax1.twiny()
            ax4 = ax2.twiny()
            # extra space for the second axis at the bottom
            fig.subplots_adjust(bottom=0.2)

            if not params.magnetized:
                for isp, sp in enumerate(params.species_names):
                    sp_vacf_str = "{} ".format(sp) + vacf_str
                    acf_avg = vacf.dataframe[(sp_vacf_str, "Total", "Mean")]
                    acf_std = vacf.dataframe[(sp_vacf_str, "Total", "Std")]

                    d_avg = coefficient[("{} Diffusion".format(sp),"Mean")]
                    d_std = coefficient[("{} Diffusion".format(sp),"Std")]

                    # Calculate axis multipliers and labels
                    xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, d_avg, "Time", "Diffusion", vacf.units)

                    ax1.plot(xmul * time, acf_avg / acf_avg[0], label=r'$D_{' + sp + '}$')
                    ax1.fill_between(
                        xmul * time,
                        (acf_avg + acf_std) / (acf_avg[0] + acf_std[0]),
                        (acf_avg - acf_std) / (acf_avg[0] - acf_std[0]), alpha=0.2)

                    ax2.plot(xmul * time, ymul * d_avg, label=r'$D_{' + sp + '}$')
                    ax2.fill_between(xmul * time, ymul * (d_avg + d_std), ymul * (d_avg - d_std), alpha=0.2)
            else:
                for isp, sp in enumerate(params.species_names):
                    par_acf_avg = vacf.dataframe["{} Parallel Velocity ACF avg".format(sp)]
                    par_acf_std = vacf.dataframe["{} Parallel Velocity ACF std".format(sp)]

                    par_d_avg = coefficient["{} Parallel Diffusion avg".format(sp)]
                    par_d_std = coefficient["{} Parallel Diffusion std".format(sp)]

                    perp_acf_avg = vacf.dataframe["{} Perpendicular Velocity ACF avg".format(sp)]
                    perp_acf_std = vacf.dataframe["{} Perpendicular Velocity ACF std".format(sp)]

                    perp_d_avg = coefficient["{} Perpendicular Diffusion avg".format(sp)]
                    perp_d_std = coefficient["{} Perpendicular Diffusion std".format(sp)]

                    # Calculate axis multipliers and labels
                    xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, perp_d_avg, "Time", "Diffusion", vacf.units)
                    # ACF Parallel plot
                    ax1.plot(xmul * time, par_acf_avg / par_acf_avg.iloc[0], label=sp + r' $Z_{\parallel}$')
                    ax1.fill_between(
                        xmul * time,
                        (par_acf_avg + par_acf_std) / (par_acf_avg.iloc[0] + par_acf_std.iloc[0]),
                        (par_acf_avg - par_acf_std) / (par_acf_avg.iloc[0] - par_acf_std.iloc[0]), alpha=0.2)
                    # Coefficient Parallel plot
                    ax2.plot(xmul * time, ymul * par_d_avg, label=sp + r' $D_{\parallel}$')
                    ax2.fill_between(xmul * time, ymul * (par_d_avg + par_d_std), ymul * (par_d_avg - par_d_std),
                                     alpha=0.2)
                    # ACF Perpendicular Plot
                    ax1.plot(xmul * time, perp_acf_avg / perp_acf_avg.iloc[0], label=sp + r' $Z_{\perp}$')
                    ax1.fill_between(
                        xmul * time,
                        (perp_acf_avg + perp_acf_std) / (perp_acf_avg.iloc[0] + perp_acf_std.iloc[0]),
                        (perp_acf_avg - perp_acf_std) / (perp_acf_avg.iloc[0] - perp_acf_std.iloc[0]), alpha=0.2)
                    # Coefficient Perpendicular Plot
                    ax2.plot(xmul * time, ymul * perp_d_avg, label=sp + r' $D_{\perp}$')
                    ax2.fill_between(xmul * time, ymul * (perp_d_avg + perp_d_std), ymul * (perp_d_avg - perp_d_std),
                                     alpha=0.2)

            xlims = (xmul * time[1], xmul * time[-1] * 1.5)

            ax1.set(xlim=xlims, xscale='log', ylabel=r'Velocity ACF', xlabel=r"Time difference" + xlbl)
            ax2.set(xlim=xlims, xscale='log', ylabel=r'Diffusion' + ylbl, xlabel=r"$\tau$" + xlbl)

            ax1.legend(loc='best')
            ax2.legend(loc='best')

            # Finish the index axes
            for axi in [ax3, ax4]:
                axi.grid(alpha=0.1)
                axi.set(xlim=[1, vacf.slice_steps * 1.5], xscale='log', xlabel='Index')

            fig.tight_layout()
            if figname:
                fig.savefig(os.path.join(vacf.saving_dir, figname))
            else:
                fig.savefig(os.path.join(vacf.saving_dir, 'Plot_Diffusion_' + vacf.job_id + '.png'))

            if show:
                fig.show()

        return coefficient

    @staticmethod
    def interdiffusion(params,
                       phase: str = 'production',
                       compute_acf: bool = True,
                       no_slices: int = 1,
                       plot: bool = True,
                       show: bool = False,
                       figname: str = None,
                       **kwargs):
        """
        Calculate the interdiffusion coefficients from the diffusion flux auto-correlation function.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters.

        phase : str, optional
            Phase to compute. Default = 'production'.

        compute_acf : bool
            Flag for recalculating the ACF. Default = True.
            If False it will read in the data from the dataframe.

        no_slices : int, optional
            Number of slices of the simulation. Default = 1.

        plot : bool, optional
            Flag to plot transport coefficient with corresponding autocorrelation function. Default = True.

        show : bool, optional
            Flag for prompting plot to screen when using IPython kernel. Default = False.

        figname : str, optional
            Name with which to save the file. It automatically saves it in the correct directory.

        **kwargs:
            Arguments to pass :meth:`sarkas.tools.observables.DiffusionFlux`

        Returns
        -------
        coefficient : pandas.DataFrame
            Pandas dataframe containing the value of the transport coefficient as a function of integration time.

        """
        print('\n\n{:=^70} \n'.format(' Interdiffusion Coefficient '))
        coefficient = pd.DataFrame()
        if compute_acf:
            jc_acf = obs.DiffusionFlux()
            jc_acf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            jc_acf.compute()

        else:
            jc_acf = obs.DiffusionFlux()
            jc_acf.setup(params, phase=phase, no_slices=no_slices, **kwargs)
            jc_acf.parse()

        # Print some info
        print('Data saved in: \n', os.path.join(jc_acf.saving_dir, 'InterDiffusion_' + jc_acf.job_id + '.h5'))

        print('\nNo. of slices = {}'.format(jc_acf.no_slices))
        print('No. dumps per slice = {}'.format(int(jc_acf.slice_steps/jc_acf.dump_step) ))

        print('Time interval of autocorrelation function = {:.4e} [s] ~ {} w_p T'.format(
            jc_acf.dt * jc_acf.slice_steps * jc_acf.dump_step,
            int(jc_acf.dt * jc_acf.slice_steps * jc_acf.dump_step * jc_acf.total_plasma_frequency)))

        no_int = jc_acf.slice_steps
        no_fluxes_acf = jc_acf.no_fluxes_acf
        # Normalization constant
        const = 1. / (3.0 * jc_acf.total_num_ptcls * jc_acf.species_concentrations.prod())

        # to_numpy creates a 2d-array, hence the [:,0]
        time = jc_acf.dataframe[("Time")].to_numpy()[:, 0]

        coefficient["Time"] = time
        df_str = "Diffusion Flux ACF"
        id_str = "Inter Diffusion Flux"
        for isl in range(jc_acf.no_slices):
            D_ij = np.zeros((no_fluxes_acf, jc_acf.slice_steps))

            for ij in range(no_fluxes_acf):
                integrand = np.array(jc_acf.dataframe[(df_str + " {}".format(ij), "Total", "slice {}".format(isl))])

                for it in range(1, no_int):
                    D_ij[ij, it] = const * np.trapz(integrand[:it], x=time[:it])

                coefficient[id_str + " {}_slice {}".format(ij, isl)] = D_ij[ij, :]

        for ij in range(no_fluxes_acf):
            col_str = [id_str + " {}_slice {}".format(ij, isl) for isl in range(jc_acf.no_slices)]
            coefficient[id_str + " {}_Mean".format(ij)] = coefficient[col_str].mean(axis=1)
            coefficient[id_str + " {}_Std".format(ij)] = coefficient[col_str].std(axis=1)

        coefficient.columns = pd.MultiIndex.from_tuples([tuple(c.split("_")) for c in coefficient.columns])
        coefficient.to_hdf(
            os.path.join(jc_acf.saving_dir, 'InterDiffusion_' + jc_acf.job_id + '.h5'),
            mode='w',
            key='interdiffusion',
            index=False)

        if plot or figname:
            # Make the plot
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 7))
            ax3 = ax1.twiny()
            ax4 = ax2.twiny()
            # extra space for the second axis at the bottom
            # fig.subplots_adjust(bottom=0.2)
            for flux in range(jc_acf.no_fluxes):
                acf_avg = jc_acf.dataframe[(df_str + " {}".format(flux), "Total", "Mean")]
                acf_std = jc_acf.dataframe[(df_str + " {}".format(flux), "Total", "Std")]

                d_avg = coefficient[(id_str + " {}".format(flux), "Mean")]
                d_std = coefficient[(id_str + " {}".format(flux), "Std")]

                # Calculate axis multipliers and labels
                xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, d_avg, "Time", "Diffusion", jc_acf.units)

                # ACF
                ax1.plot(xmul * time, acf_avg / acf_avg.iloc[0], label=r'Flux $J_{}$'.format(flux))
                ax1.fill_between(
                    xmul * time,
                    (acf_avg + acf_std) / (acf_avg.iloc[0] + acf_std.iloc[0]),
                    (acf_avg - acf_std) / (acf_avg.iloc[0] - acf_std.iloc[0]), alpha=0.2)

                # Coefficient
                ax2.plot(xmul * time, ymul * d_avg, label=r'IDF $D_{}$'.format(flux))
                ax2.fill_between(xmul * time, ymul * (d_avg + d_std), ymul * (d_avg - d_std), alpha=0.2)

            xlims = (xmul * time[1], xmul * time[-1] * 1.5)

            ax1.set(xlim=xlims, xscale='log', ylabel=r'Diffusion Flux ACF', xlabel=r"Time difference" + xlbl)
            ax2.set(xlim=xlims, xscale='log', ylabel=r'Inter Diffusion' + ylbl, xlabel=r"$\tau$" + xlbl)

            ax1.legend(loc='best')
            ax2.legend(loc='best')
            # Finish the index axes
            for axi in [ax3, ax4]:
                axi.grid(alpha=0.1)
                axi.set(xlim=(1, jc_acf.slice_steps * 1.5), xscale='log', xlabel='Index')

            fig.tight_layout()
            if figname:
                fig.savefig(os.path.join(jc_acf.saving_dir, figname))
            else:
                fig.savefig(os.path.join(jc_acf.saving_dir, 'Plot_InterDiffusion_' + jc_acf.job_id + '.png'))

            if show:
                fig.show()

        # Save to dataframe

        return coefficient

    @staticmethod
    def viscosity(params, phase: str = 'production', show=False):
        """
        TODO: Calculate bulk and shear viscosity from pressure auto-correlation function.

        Parameters
        ----------
        params : sarkas.core.Parameters
            Simulation's parameters.

        phase : str
            Phase to analyze. Default = 'production'.

        show : bool
            Flag for prompting plot to screen.

        Returns
        -------
        coefficient : pandas.DataFrame
            Pandas dataframe containing the value of the transport coefficient as a function of integration time.

        """
        print('\n\n{:=^70} \n'.format(' Viscosity Coefficient '))
        coefficient = pd.DataFrame()
        energies = obs.Thermodynamics()
        energies.setup(params, phase)
        energies.parse('production')
        beta = (energies.kB * energies.dataframe["Temperature"].mean()) ** (-1.0)
        time = np.array(energies.dataframe["Time"])
        coefficient["Time"] = time
        dim_lbl = ['x', 'y', 'z']
        shear_viscosity = np.zeros((params.dimensions, params.dimensions, energies.prod_no_dumps))
        bulk_viscosity = np.zeros(energies.prod_no_dumps)
        const = params.box_volume * beta
        if not 'Pressure Tensor ACF xy' in energies.dataframe.columns:
            print('Pressure not yet calculated')
            print("Calculating Pressure quantities ...")
            energies.compute_pressure_quantities()

        # Calculate the acf of the pressure tensor

        fig, axes = plt.subplots(2, 3, sharex=True, figsize=(16, 10))
        for i, ax1 in enumerate(dim_lbl):
            for j, ax2 in enumerate(dim_lbl):
                integrand = np.array(energies.dataframe["Pressure Tensor ACF {}{}".format(ax1, ax2)])
                for it in range(1, energies.prod_no_dumps):
                    shear_viscosity[i, j, it] = const * np.trapz(integrand[:it], x=time[:it])

                coefficient["{}{} Shear Viscosity Tensor".format(ax1, ax2)] = shear_viscosity[i, j, :]
                xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, shear_viscosity[i, j, :],
                                                               "Time", "Viscosity", energies.units)
                if "{}{}".format(ax1, ax2) in ["yx", "xy"]:
                    axes[0, 0].semilogx(xmul * time, integrand / integrand[0],
                                        label=r"$P_{" + "{}{}".format(ax1, ax2) + " }(t)$")
                    axes[1, 0].semilogx(xmul * time, ymul * shear_viscosity[i, j, :],
                                        label=r"$\eta_{ " + "{}{}".format(ax1, ax2) + " }(t)$")

                elif "{}{}".format(ax1, ax2) in ["xz", "zx"]:
                    axes[0, 1].semilogx(xmul * time, integrand / integrand[0],
                                        label=r"$P_{" + "{}{}".format(ax1, ax2) + " }(t)$")
                    axes[1, 1].semilogx(xmul * time, ymul * shear_viscosity[i, j, :],
                                        label=r"$\eta_{ " + "{}{}".format(ax1, ax2) + " }(t)$")

                elif "{}{}".format(ax1, ax2) in ["yz", "zy"]:
                    axes[0, 2].semilogx(xmul * time, integrand / integrand[0],
                                        label=r"$P_{" + "{}{}".format(ax1, ax2) + " }(t)$")
                    axes[1, 2].semilogx(xmul * time, ymul * shear_viscosity[i, j, :],
                                        label=r"$\eta_{ " + "{}{}".format(ax1, ax2) + " }(t)$")
            axes[0, 0].legend()
            axes[0, 1].legend()
            axes[0, 2].legend()
            axes[1, 0].legend()
            axes[1, 1].legend()
            axes[1, 2].legend()
            axes[0, 0].set_xlabel(r"Time " + xlbl)
            axes[0, 1].set_xlabel(r"Time " + xlbl)
            axes[0, 2].set_xlabel(r"Time " + xlbl)

            axes[1, 0].set_xlabel(r"Time " + xlbl)
            axes[1, 1].set_xlabel(r"Time " + xlbl)
            axes[1, 2].set_xlabel(r"Time " + xlbl)

            axes[0, 0].set_ylabel(r"Pressure Tensor ACF")
            axes[1, 0].set_ylabel(r"Shear Viscosity" + ylbl)

            fig.tight_layout()
            fig.savefig(os.path.join(energies.fldr, "ShearViscosity_Plots_" + energies.job_id + ".png"))
            if show:
                fig.show()

        # Calculate Bulk Viscosity
        pressure_acf = obs.autocorrelationfunction_1D(np.array(energies.dataframe["Pressure"])
                                                      - energies.dataframe["Pressure"].mean())
        bulk_integrand = pressure_acf
        for it in range(1, energies.prod_no_dumps):
            bulk_viscosity[it] = const * np.trapz(bulk_integrand[:it], x=time[:it])

        coefficient["Bulk Viscosity"] = bulk_viscosity

        fig, ax = plt.subplots(2, 1)
        xmul, ymul, _, _, xlbl, ylbl = obs.plot_labels(time, bulk_viscosity, "Time", "Viscosity", energies.units)
        ax[0].semilogx(xmul * time, bulk_integrand / bulk_integrand[0], label=r"$P(t)$")
        ax[1].semilogx(xmul * time, ymul * bulk_viscosity, label=r"$\eta_{V}(t)$")

        ax[0].set_xlabel(r"Time" + xlbl)
        ax[1].set_xlabel(r"Dumps")
        ax[0].set_ylabel(r"Pressure ACF")
        ax[1].set_ylabel(r"Bulk Viscosity" + ylbl)
        ax[0].legend()
        ax[1].legend()
        fig.tight_layout()
        fig.savefig(os.path.join(energies.fldr, "BulkViscosity_Plots_" + energies.job_id + ".png"))
        if show:
            fig.show()

        return coefficient



"""
Module for handling Exact Gradient corrected Screened (EGS) Potential
"""
import numpy as np
from numba import njit
import fdint


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params: sarkas.core.Parameters
        Simulation's parameters.

    """

    twopi = 2.0 * np.pi

    # lambda factor : 1 = von Weizsaecker, 1/9 = Thomas-Fermi
    if not hasattr(potential, 'lmbda'):
        potential.lmbda = 1.0 / 9.0

    fdint_dfdk_vec = np.vectorize(fdint.dfdk)
    # eq. (14) of Ref. [1]_
    params.nu = 3.0/np.pi ** 1.5 * params.landau_length / params.lambda_deB
    params.nu *= potential.lmbda * fdint_dfdk_vec(k=-0.5, phi=params.eta_e)

    # Degeneracy Parameter
    theta = params.electron_degeneracy_parameter
    if 0.1 <= theta <= 12:
        # Regime of validity of the following approximation Perrot et al. Phys Rev A 302619 (1984)
        # eq. (33) of Ref. [1]_
        Ntheta = 1.0 + 2.8343 * theta ** 2 - 0.2151 * theta ** 3 + 5.2759 * theta ** 4
        # eq. (34) of Ref. [1]_
        Dtheta = 1.0 + 3.9431 * theta ** 2 + 7.9138 * theta ** 4
        # eq. (32) of Ref. [1]_
        h = Ntheta / Dtheta * np.tanh(1.0 / theta)
        # grad h(x)
        gradh = (-(Ntheta / Dtheta) / np.cosh(1 / theta) ** 2 / (theta ** 2)  # derivative of tanh(1/x)
                 - np.tanh(1.0 / theta) * (
                         Ntheta * (7.8862 * theta + 31.6552 * theta ** 3) / Dtheta ** 2  # derivative of 1/Dtheta
                         + (5.6686 * theta - 0.6453 * theta ** 2 + 21.1036 * theta ** 3) / Dtheta) # derivative of Ntheta
                 )
        # eq.(31) of Ref. [1]_
        b = 1.0 - 2.0 / ( 8.0 * (params.kF * params.lambda_TF)**2) * (h - 2.0 * theta * gradh)
    else:
        b = 1.0

    params.b = b

    # Monotonic decay
    if params.nu <= 1:
        # eq. (29) of Ref. [1]_
        params.lambda_p = params.lambda_TF * np.sqrt(params.nu / (2.0 * b + 2.0 * np.sqrt(b ** 2 - params.nu)))
        params.lambda_m = params.lambda_TF * np.sqrt(params.nu / (2.0 * b - 2.0 * np.sqrt(b ** 2 - params.nu)))
        params.alpha = b / np.sqrt(b - params.nu)

    # Oscillatory behavior
    if params.nu > 1:
        # eq. (29) of Ref. [1]_
        params.gamma_m = params.lambda_TF * np.sqrt(params.nu / (np.sqrt(params.nu) - b))
        params.gamma_p = params.lambda_TF * np.sqrt(params.nu / (np.sqrt(params.nu) + b))
        params.alphap = b / np.sqrt(params.nu - b)

    potential.matrix = np.zeros((6, params.num_species, params.num_species))

    potential.matrix[1, :, :] = params.nu

    for i, q1 in enumerate(params.species_charges):

        for j, q2 in enumerate(params.species_charges):

            if params.nu <= 1:
                potential.matrix[0, i, j] = q1 * q2 / (2.0 * params.fourpie0)
                potential.matrix[2, i, j] = (1.0 + params.alpha)
                potential.matrix[3, i, j] = (1.0 - params.alpha)
                potential.matrix[4, i, j] = 1.0 / params.lambda_m
                potential.matrix[5, i, j] = 1.0 / params.lambda_p

            if params.nu > 1:
                potential.matrix[0, i, j] = q1 * q2 / params.fourpie0
                potential.matrix[2, i, j] = 1.0
                potential.matrix[3, i, j] = params.alphap
                potential.matrix[4, i, j] = 1.0 / params.gamma_m
                potential.matrix[5, i, j] = 1.0 / params.gamma_p

    assert potential.method == "PP", "P3M Algorithm not implemented yet. Good Bye!"

    potential.force = egs_force
    params.force_error = np.sqrt(twopi / params.lambda_TF) * np.exp(-potential.rc / params.lambda_TF)
    # Renormalize
    params.force_error *= params.a_ws ** 2 * np.sqrt(params.total_num_ptcls / params.box_volume)


@njit
def egs_force(r, pot_matrix):
    """ 
    Calculates Potential and force between particles using the EGS Potential.
    
    Parameters
    ----------
    r : float
        Particles' distance.

    pot_matrix : array
        EGS potential parameters. 

    Return
    ------

    U : float
        Potential.

    fr : float
        Force.

    """
    # nu = pot_matrix[1]
    if pot_matrix[1] <= 1.0:
        # pot_matrix[0] = Charge factor = q^2/4pi eps0 if mks q^2 if cgs
        # pot_matrix[2] = 1 + alpha
        # pot_matrix[3] = 1 - alpha
        # pot_matrix[4] = 1.0 / lambda_minus
        # pot_matrix[5] = 1.0 / lambda_plus

        temp1 = pot_matrix[2] * np.exp(-r * pot_matrix[4])
        temp2 = pot_matrix[3] * np.exp(-r * pot_matrix[5])
        # Potential
        U = (temp1 + temp2) * pot_matrix[0] / r
        # Force
        fr = U / r + pot_matrix[0] * (temp1 * pot_matrix[4] + temp2 * pot_matrix[5]) / r

    else:
        # pot_matrix[0] = Charge factor
        # pot_matrix[2] = 1.0
        # pot_matrix[3] = alpha prime
        # pot_matrix[4] = 1.0 / gamma_minus
        # pot_matrix[5] = 1.0 / gamma_plus
        cos = np.cos(r * pot_matrix[4])
        sin = np.sin(r * pot_matrix[4])
        exp = pot_matrix[0] * np.exp(-r * pot_matrix[5])
        U = (cos + pot_matrix[3] * sin) * exp / r
        fr = U / r  # derivative of 1/r
        fr += U * pot_matrix[5]  # derivative of exp
        fr += pot_matrix[4] * (sin - pot_matrix[3] * cos) * exp / r

    return U, fr
